@@ -241,9 +241,18 @@ typedef struct {
    Time		freeTime;
    int		lastKnownPitch;
 
-   int          size;
-   void         *video_memory;
-   int          video_offset;
+   void         *video_memory;			/* opaque memory management information structure */
+   CARD32       video_offset;			/* offset in video memory of packed YUV buffer */
+
+   void         *video_planarmem;		/* opaque memory management information structure */
+   CARD32       video_planarbuf; 		/* offset in video memory of planar YV12 buffer */
+   
+#ifdef XF86DRI
+   Bool         tried_agp;			/* TRUE if AGP allocation has been tried */
+   CARD32	agpBase;			/* Physical address of aperture base */
+   CARD32	agpBufferOffset;		/* Offset of buffer in AGP memory, or 0 if unavailable */
+   drmAddress   agpBufferMap;			/* Mapping of AGP buffer in process memory, or NULL */
+#endif
 
 } SavagePortPrivRec, *SavagePortPrivPtr;
 
@@ -375,7 +384,7 @@ void SavageInitVideo(ScreenPtr pScreen)
             adaptors = &newAdaptor;
         } else {
             newAdaptors =  /* need to free this someplace */
-        	xalloc((num_adaptors + 1) * sizeof(XF86VideoAdaptorPtr*));
+        	malloc((num_adaptors + 1) * sizeof(XF86VideoAdaptorPtr*));
             if(newAdaptors) {
         	memcpy(newAdaptors, adaptors, num_adaptors * 
         				sizeof(XF86VideoAdaptorPtr));
@@ -390,7 +399,7 @@ void SavageInitVideo(ScreenPtr pScreen)
         xf86XVScreenInit(pScreen, adaptors, num_adaptors);
 
     if(newAdaptors)
-	xfree(newAdaptors);
+	free(newAdaptors);
 
     if( newAdaptor )
     {
@@ -875,7 +884,7 @@ SavageSetupImageVideo(ScreenPtr pScreen)
 
     xf86ErrorFVerb(XVTRACE,"SavageSetupImageVideo\n");
 
-    if(!(adapt = xcalloc(1, sizeof(XF86VideoAdaptorRec) +
+    if(!(adapt = calloc(1, sizeof(XF86VideoAdaptorRec) +
 			    sizeof(SavagePortPrivRec) +
 			    sizeof(DevUnion))))
 	return NULL;
@@ -1028,7 +1037,7 @@ static void
 SavageStopVideo(ScrnInfoPtr pScrn, pointer data, Bool shutdown)
 {
     SavagePortPrivPtr pPriv = (SavagePortPrivPtr)data;
-    /*SavagePtr psav = SAVPTR(pScrn); */
+    SavagePtr psav = SAVPTR(pScrn);
 
     xf86ErrorFVerb(XVTRACE,"SavageStopVideo\n");
 
@@ -1037,10 +1046,30 @@ SavageStopVideo(ScrnInfoPtr pScrn, pointer data, Bool shutdown)
     if(shutdown) {
       /*SavageClipVWindow(pScrn);*/
  	SavageStreamsOff( pScrn );
+
+#ifdef XF86DRI
+	if (pPriv->agpBufferMap != NULL) {
+	    SAVAGEDRIServerPrivatePtr pSAVAGEDRIServer = psav->DRIServerInfo;
+
+            /* agpXVideo is reused to implement UploadToScreen in EXA */            
+            if (!psav->useEXA) {
+	        drmUnmap(pPriv->agpBufferMap, pSAVAGEDRIServer->agpXVideo.size);
+	        pSAVAGEDRIServer->agpXVideo.map = NULL;
+            }
+	    pPriv->agpBufferMap = NULL;
+	    pPriv->agpBufferOffset = 0;
+	}
+	pPriv->tried_agp = FALSE;
+#endif
+
         if (pPriv->video_memory != NULL) {
 	    SavageFreeMemory(pScrn, pPriv->video_memory);
 	    pPriv->video_memory = NULL;
         }
+        if (pPriv->video_planarmem != NULL) {
+	    SavageFreeMemory(pScrn, pPriv->video_planarmem);
+	    pPriv->video_planarmem = NULL;
+        }
 	pPriv->videoStatus = 0;
     } else {
 	if(pPriv->videoStatus & CLIENT_VIDEO_ON) {
@@ -1170,34 +1199,36 @@ SavageCopyPlanarDataBCI(
     unsigned char *srcV, /* V */
     unsigned char *srcU, /* U */
     unsigned char *dst,
+    unsigned char * planarPtr,
+    unsigned long planarOffset,
     int srcPitch, int srcPitch2,
     int dstPitch,
-    int h,int w)
+    int h,int w,
+    Bool isAGP)
 {
     SavagePtr psav = SAVPTR(pScrn);
-    /* half of the dest buffer for copying the YVU data to it ??? */
-    unsigned char *dstCopy = (unsigned char *)(((unsigned long)dst
-                                                + dstPitch * h
-                                                + 0x0f) & ~0x0f);
+
     /* for pixel transfer */
-    unsigned long offsetY = (unsigned long)dstCopy - (unsigned long)psav->FBBase;
+    unsigned long offsetY = planarOffset;
     unsigned long offsetV = offsetY +  srcPitch * h;
     unsigned long offsetU = offsetV +  srcPitch2 * (h>>1);
     unsigned long dstOffset  = (unsigned long)dst - (unsigned long)psav->FBBase;
-    int i;
+    unsigned char memType;
     
     BCI_GET_PTR;
 
     /* copy Y planar */
-    memcpy(dstCopy, srcY, srcPitch * h);
+    memcpy(planarPtr, srcY, srcPitch * h);
 
     /* copy V planar */    
-    dstCopy = dstCopy + srcPitch * h;
-    memcpy(dstCopy, srcV, srcPitch2 * (h>>1));
+    planarPtr = planarPtr + srcPitch * h;
+    memcpy(planarPtr, srcV, srcPitch2 * (h>>1));
 
     /* copy U planar */
-    dstCopy = dstCopy + srcPitch2 * (h>>1);    
-    memcpy(dstCopy, srcU, srcPitch2 * (h>>1));
+    planarPtr = planarPtr + srcPitch2 * (h>>1);    
+    memcpy(planarPtr, srcU, srcPitch2 * (h>>1));
+
+    memType = isAGP ? 3 : 0;
 
     /*
      * Transfer pixel data from one memory location to another location
@@ -1217,23 +1248,18 @@ SavageCopyPlanarDataBCI(
 
     w = (w+0xf)&0xff0;
     psav->WaitQueue(psav,11);
-    BCI_SEND(0x96070051);
-    BCI_SEND(offsetY);
-
+    BCI_SEND(BCI_SET_REGISTER | BCI_SET_REGISTER_COUNT(7) | 0x51);
+    BCI_SEND(offsetY | memType);
     BCI_SEND(dstOffset);
-
     BCI_SEND(((h-1)<<16)|((w-1)>>3));
-
     BCI_SEND(dstPitch >> 3);
-
-
-    BCI_SEND(offsetU);
-    BCI_SEND(offsetV);
-
+    BCI_SEND(offsetU | memType);
+    BCI_SEND(offsetV | memType);
     BCI_SEND((srcPitch2 << 16)| srcPitch2);
 
-    BCI_SEND(0x96010050);
+    BCI_SEND(BCI_SET_REGISTER | BCI_SET_REGISTER_COUNT(1) | 0x50);
     BCI_SEND(0x00200003 | srcPitch);
+
     BCI_SEND(0xC0170000);
 }
 
@@ -1304,6 +1330,8 @@ SavageVideoSave(ScreenPtr pScreen, ExaOffscreenArea *area)
 
     if (pPriv->video_memory == area)
         pPriv->video_memory = NULL;
+    if (pPriv->video_planarmem == area)
+        pPriv->video_planarmem = NULL;
 }
 
 static CARD32
@@ -1475,14 +1503,17 @@ SavageDisplayVideoOld(
       }
 
     if (S3_MOBILE_TWISTER_SERIES(psav->Chipset)
-        && psav->FPExpansion) {
-        drw_w = (((float)(drw_w * psav->XExp1)/(float)psav->XExp2)+1);
-        drw_h = (float)(drw_h * psav->YExp1)/(float)psav->YExp2+1;
-        dstBox->x1 = (float)(dstBox->x1 * psav->XExp1)/(float)psav->XExp2;
-        dstBox->y1 = (float)(dstBox->y1 * psav->YExp1)/(float)psav->YExp2;
-
-        dstBox->x1 += psav->displayXoffset;
-        dstBox->y1 += psav->displayYoffset;
+	&& psav->FPExpansion) {
+	drw_w = (drw_w * psav->XExp1) / psav->XExp2 + 1;
+	drw_h = (drw_h * psav->YExp1) / psav->YExp2 + 1;
+	dstBox->x1 = (dstBox->x1 * psav->XExp1) / psav->XExp2;
+	dstBox->y1 = (dstBox->y1 * psav->YExp1) / psav->YExp2;
+	dstBox->x2 = (dstBox->x2 * psav->XExp1) / psav->XExp2;
+	dstBox->y2 = (dstBox->y2 * psav->YExp1) / psav->YExp2;
+	dstBox->x1 += psav->displayXoffset;
+	dstBox->y1 += psav->displayYoffset;
+	dstBox->x2 += psav->displayXoffset;
+	dstBox->y2 += psav->displayYoffset;
     }
 
     /*
@@ -1636,12 +1667,16 @@ SavageDisplayVideoNew(
 	    !psav->CrtOnly &&
 	    !psav->TvOn) 
     {
-	drw_w = (drw_w * psav->XExp1)/psav->XExp2 + 1;
-	drw_h = (drw_h * psav->YExp1)/psav->YExp2 + 1;
-	dstBox->x1 = (dstBox->x1 * psav->XExp1)/psav->XExp2;
-	dstBox->y1 = (dstBox->y1 * psav->YExp1)/psav->YExp2;
+	drw_w = (drw_w * psav->XExp1) / psav->XExp2 + 1;
+	drw_h = (drw_h * psav->YExp1) / psav->YExp2 + 1;
+	dstBox->x1 = (dstBox->x1 * psav->XExp1) / psav->XExp2;
+	dstBox->y1 = (dstBox->y1 * psav->YExp1) / psav->YExp2;
+	dstBox->x2 = (dstBox->x2 * psav->XExp1) / psav->XExp2;
+	dstBox->y2 = (dstBox->y2 * psav->YExp1) / psav->YExp2;
 	dstBox->x1 += psav->displayXoffset;
 	dstBox->y1 += psav->displayYoffset;
+	dstBox->x2 += psav->displayXoffset;
+	dstBox->y2 += psav->displayYoffset;
     }
 
 	if (psav->IsSecondary) {
@@ -1840,6 +1875,16 @@ SavageDisplayVideo2000(
 #endif
 }
 
+static void
+SavageFillKeyHelper(DrawablePtr pDraw, uint32_t colorKey, RegionPtr clipBoxes)
+{
+#if HAVE_XV_DRAWABLE_HELPER
+    xf86XVFillKeyHelperDrawable(pDraw, colorKey, clipBoxes);
+#else
+    xf86XVFillKeyHelper(pDraw->pScreen, colorKey, clipBoxes);
+#endif
+}
+
 static int 
 SavagePutImage( 
     ScrnInfoPtr pScrn, 
@@ -1860,6 +1905,7 @@ SavagePutImage(
     unsigned char *dst_start;
     int pitch, new_size, offset, offsetV=0, offsetU=0;
     int srcPitch, srcPitch2=0, dstPitch;
+    int planarFrameSize;
     int top, left, npixels, nlines;
     BoxRec dstBox;
     CARD32 tmp;
@@ -1895,8 +1941,8 @@ SavagePutImage(
 
     pitch = pScrn->bitsPerPixel * pScrn->displayWidth >> 3;
 
+    /* All formats directly displayable by Savage are packed and 2 bytes per pixel */
     dstPitch = ((width << 1) + 15) & ~15;
-    /*new_h = ((dstPitch * height) + pitch - 1) / pitch;*/
     new_size = dstPitch * height;
 
     switch(id) {
@@ -1923,17 +1969,79 @@ SavagePutImage(
 	break;
     }  
 
+    /* Calculate required memory for all planar frames */
+    planarFrameSize = 0;
     if (srcPitch2 != 0 && S3_SAVAGE4_SERIES(psav->Chipset) && psav->BCIforXv) {
-        new_size = ((new_size + 0xF) & ~0xF) + srcPitch * height + srcPitch2 * height;
+        new_size = ((new_size + 0xF) & ~0xF);
+        planarFrameSize = srcPitch * height + srcPitch2 * height;
     }
 
-/*    if(!(pPriv->area = SavageAllocateMemory(pScrn, pPriv->area, new_h)))
-	return BadAlloc;*/
-    pPriv->video_offset = SavageAllocateMemory(pScrn, &pPriv->video_memory,
-					      new_size);
+    /* Check whether AGP buffers can be allocated. If not, fall back to ordinary
+       upload to framebuffer (slower) */
+#ifdef XF86DRI
+    if (!pPriv->tried_agp && !psav->IsPCI && psav->drmFD > 0 && psav->DRIServerInfo != NULL) {
+	SAVAGEDRIServerPrivatePtr pSAVAGEDRIServer = psav->DRIServerInfo;
+        
+	pPriv->tried_agp = TRUE;
+	if (pSAVAGEDRIServer->agpXVideo.size >= max(new_size, planarFrameSize)) {
+	    if (pSAVAGEDRIServer->agpXVideo.map == NULL &&
+	        drmMap( psav->drmFD,
+		pSAVAGEDRIServer->agpXVideo.handle,
+		pSAVAGEDRIServer->agpXVideo.size,
+		&pSAVAGEDRIServer->agpXVideo.map ) < 0 ) {
+
+		xf86DrvMsg( pScreen->myNum, X_ERROR, "[agp] XVideo: Could not map agpXVideo \n" );
+		pPriv->agpBufferOffset = 0;
+		pPriv->agpBufferMap = NULL;
+	    } else {
+		pPriv->agpBufferMap = pSAVAGEDRIServer->agpXVideo.map;
+		pPriv->agpBufferOffset = pSAVAGEDRIServer->agpXVideo.offset;
+		pPriv->agpBase = drmAgpBase(psav->drmFD);
+#if 0
+		xf86DrvMsg( pScreen->myNum, X_INFO,
+		       "[agp] agpXVideo mapped at 0x%08lx aperture=0x%08x offset=0x%08lx\n",
+		       (unsigned long)pPriv->agpBufferMap, pPriv->agpBase, pPriv->agpBufferOffset);
+#endif
+	    }
+	} else {
+	    /* This situation is expected if AGPforXv is disabled, otherwise report. */
+	    if (pSAVAGEDRIServer->agpXVideo.size > 0) {
+		xf86DrvMsg( pScreen->myNum, X_ERROR,
+		    "[agp] XVideo: not enough space in buffer (got %ld bytes, required %d bytes).\n", 
+	    	    (long int)pSAVAGEDRIServer->agpXVideo.size, max(new_size, planarFrameSize));
+	    }
+	    pPriv->agpBufferMap = NULL;
+	    pPriv->agpBufferOffset = 0;
+	}
+    }
+#endif /* XF86DRI */
+
+
+    /* Buffer for final packed frame */
+    pPriv->video_offset = SavageAllocateMemory(
+	pScrn, &pPriv->video_memory,
+	new_size);
     if (pPriv->video_offset == 0)
         return BadAlloc;
 
+    /* Packed format cases */
+    if (planarFrameSize == 0) {
+	pPriv->video_planarbuf = 0;
+
+    /* Planar format cases */
+    } else {
+	/* Hardware-assisted planar conversion only works on 16-byte aligned addresses */
+	pPriv->video_planarbuf = SavageAllocateMemory(
+	    pScrn, &pPriv->video_planarmem,
+	    ((planarFrameSize + 0xF) & ~0xF));
+	if (pPriv->video_planarbuf != 0) {
+	    /* TODO: stop any pending conversions when buffers change... */
+	    pPriv->video_planarbuf = ((pPriv->video_planarbuf + 0xF) & ~0xF);
+	} else {
+	    /* Fallback using software conversion */
+	}
+    }
+
     /* copy data */
     top = y1 >> 16;
     left = (x1 >> 16) & ~1;
@@ -1941,7 +2049,6 @@ SavagePutImage(
     left <<= 1;
 
     offset = (pPriv->video_offset) + (top * dstPitch);
-    /*offset = pPriv->area->box.y1 * psav->lDelta;*/
     dst_start = (psav->FBBase + ((offset + left) & ~BASE_PAD));
 
     switch(id) {
@@ -1952,13 +2059,33 @@ SavagePutImage(
 	offsetU += tmp;
 	offsetV += tmp;
 	nlines = ((((y2 + 0xffff) >> 16) + 1) & ~1) - top;
-        if (S3_SAVAGE4_SERIES(psav->Chipset) && psav->BCIforXv && (npixels & 0xF) == 0) {
-            SavageCopyPlanarDataBCI(
-                pScrn,
-	    	buf + (top * srcPitch) + (left >> 1), 
-	    	buf + offsetV, 
-	    	buf + offsetU, 
-	    	dst_start, srcPitch, srcPitch2, dstPitch, nlines, npixels);
+        if (S3_SAVAGE4_SERIES(psav->Chipset) && psav->BCIforXv && (npixels & 0xF) == 0 && pPriv->video_planarbuf != 0) {
+#ifdef XF86DRI
+            if (pPriv->agpBufferMap != NULL) {
+		/* Using copy to AGP memory */
+		SavageCopyPlanarDataBCI(
+		    pScrn,
+		    buf + (top * srcPitch) + (left >> 1), 
+		    buf + offsetV, 
+		    buf + offsetU, 
+		    dst_start,
+		    pPriv->agpBufferMap,
+		    pPriv->agpBase + pPriv->agpBufferOffset,
+		    srcPitch, srcPitch2, dstPitch, nlines, npixels, TRUE);
+            } else
+#endif /* XF86DRI */
+            {
+		/* Using ordinary copy to framebuffer */
+		SavageCopyPlanarDataBCI(
+		    pScrn,
+		    buf + (top * srcPitch) + (left >> 1), 
+		    buf + offsetV, 
+		    buf + offsetU, 
+		    dst_start,
+		    (unsigned char *)psav->FBBase + pPriv->video_planarbuf,
+		    pPriv->video_planarbuf,
+		    srcPitch, srcPitch2, dstPitch, nlines, npixels, FALSE);
+	    }
         } else {
 	    SavageCopyPlanarData(
 	    	buf + (top * srcPitch) + (left >> 1), 
@@ -1988,7 +2115,7 @@ SavagePutImage(
     if(!REGION_EQUAL(pScreen, &pPriv->clip, clipBoxes)) {
 	REGION_COPY(pScreen, &pPriv->clip, clipBoxes);
 	/* draw these */
-	xf86XVFillKeyHelper(pScrn->pScreen, pPriv->colorKey, clipBoxes);
+	SavageFillKeyHelper(pDraw, pPriv->colorKey, clipBoxes);
 
     }
 
@@ -2085,18 +2212,18 @@ SavageAllocateSurface(
     surface->width = w;
     surface->height = h;
 
-    if(!(surface->pitches = xalloc(sizeof(int)))) {
+    if(!(surface->pitches = malloc(sizeof(int)))) {
 	SavageFreeMemory(pScrn, surface_memory);
 	return BadAlloc;
     }
-    if(!(surface->offsets = xalloc(sizeof(int)))) {
-	xfree(surface->pitches);
+    if(!(surface->offsets = malloc(sizeof(int)))) {
+	free(surface->pitches);
 	SavageFreeMemory(pScrn, surface_memory);
 	return BadAlloc;
     }
-    if(!(pPriv = xalloc(sizeof(OffscreenPrivRec)))) {
-	xfree(surface->pitches);
-	xfree(surface->offsets);
+    if(!(pPriv = malloc(sizeof(OffscreenPrivRec)))) {
+	free(surface->pitches);
+	free(surface->offsets);
 	SavageFreeMemory(pScrn, surface_memory);
 	return BadAlloc;
     }
@@ -2141,9 +2268,9 @@ SavageFreeSurface(
     if(pPriv->isOn)
 	SavageStopSurface(surface);
     SavageFreeMemory(pScrn, pPriv->surface_memory);
-    xfree(surface->pitches);
-    xfree(surface->offsets);
-    xfree(surface->devPrivate.ptr);
+    free(surface->pitches);
+    free(surface->offsets);
+    free(surface->devPrivate.ptr);
 
     return Success;
 }
@@ -2212,7 +2339,7 @@ SavageDisplaySurface(
 	     surface->width, surface->height, surface->pitches[0],
 	     x1, y1, x2, y2, &dstBox, src_w, src_h, drw_w, drw_h);
 
-    xf86XVFillKeyHelper(pScrn->pScreen, portPriv->colorKey, clipBoxes);
+    xf86XVFillKeyHelper(pScreen, portPriv->colorKey, clipBoxes);
 
     pPriv->isOn = TRUE;
 #if 0
@@ -2236,7 +2363,7 @@ SavageInitOffscreenImages(ScreenPtr pScreen)
 
     /* need to free this someplace */
     if (!psav->offscreenImages) {
-	if(!(offscreenImages = xalloc(sizeof(XF86OffscreenImageRec))))
+	if(!(offscreenImages = malloc(sizeof(XF86OffscreenImageRec))))
 	    return;
 	psav->offscreenImages = offscreenImages;
     } else {
